#include <gint/display.h>
#include <gint/keyboard.h>
#include <gint/timer.h>
#include <gint/gint.h>
#include <gint/cpu.h>
#include <gint/dma.h>
#include <gint/drivers/r61524.h>

#include <fxlibc/printf.h>
#include <string.h>
#include <stdlib.h>

#include "duet.h"

/* Game state */
typedef enum {
    State_Playing,
    State_Dead,
    State_Rewind,
    State_Transition,
    State_EpisodeTransition,
    State_Finale,
} state_t;

/* All levels, classified by episode */
episode_t episodes[] = {
    {
        .name = "ignorance",
        .level_count = 5,
        .levels = (level_t *[]){
            &level1, &level2, &level3, &level4, &level5
        },
    },
    {
        .name = "denial",
        .level_count = 6,
        .levels = (level_t *[]){
            &level6, &level7, &level8, &level9, &level10,
            &level11,
        },
    },
    {
        .name = "anger",
        .level_count = 5,
        .levels = (level_t *[]){
            &level12, &level13, &level14, &level15, &level16,
        },
    },
    {
        .name = "bargaining",
        .level_count = 6,
        .levels = (level_t *[]){
            &level17, &level18, &level19, &level20, &level21,
            &level22,
        },
    },
};
int episode_count = 4;

void load_level(game_t *game, level_t const *lv)
{
    game->level = lv;
    game->time = -11.0;

    game->rects = realloc(game->rects, lv->block_count * sizeof *game->rects);
    if(!game->rects)
        exit(1);
    for(int i = 0; i < lv->block_count; i++) {
        game->rects[i].meta = &lv->blocks[i];
        rect_load(&game->rects[i], &lv->blocks[i]);
    }
    game->rect_count = lv->block_count;
}

int strcount(char const *str, int c)
{
    int count = 0;
    for(int i = 0; str[i]; i++)
        count += (str[i] == c);
    return count;
}

void game_loop(int current_episode, int current_level)
{
    game_t game;
    memset(&game, 0x00, sizeof game);

    load_level(&game, episodes[current_episode].levels[current_level]);

    state_t state = State_Playing;

    if(current_level == 0) {
        state = State_EpisodeTransition;
        game.time_episode_transition = 0;
        game.time = -14.0;
    }

    /* Direction of time (with a factor regulating game speed) */
    float const time_direction_forward = 2.7;
    float const time_direction_rewind = -25;

    volatile int need_frame = 1;
    int timer = timer_configure(TIMER_ANY, 33000, GINT_CALL_SET(&need_frame));
    if(timer >= 0) timer_start(timer);

    while (1) {
        level_t const *lv = game.level;

        while (need_frame == 0) sleep();
        need_frame = 0;

        /* Time management */

        float time_direction = lv->tempo * time_direction_forward;
        if(state == State_Rewind)
            time_direction = time_direction_rewind;

        if(state == State_Dead) {
            game.time_dead += (1.0 / 30);
            if(game.time_dead >= 0.7) {
                game.time_dead = 0;
                game.forced_player_rota = game.player_rota / (game.time + 5.0);
                state = State_Rewind;
            }
            continue;
        }
        if(state == State_Transition) {
            time_direction = 1.0;
            game.time_transition += (1.0 / 30);
            if(game.time_transition >= 1.0) {
                game.time_transition = 0;
                game.player_rota = 0;
                game.forced_player_rota = 0;
                state = State_Playing;
            }
        }
        if(state == State_EpisodeTransition) {
            time_direction = 1.0;
            game.time_episode_transition += (1.0 / 30);
            if(game.time_episode_transition >= 1.0) {
                game.player_rota = 0;
                game.forced_player_rota = 0;
            }
            if(game.time_episode_transition >= 3.0) {
                game.time_episode_transition = 0;
                state = State_Playing;
            }
        }

        float dt = (1.0 / 30) * time_direction;
        game.time += dt;

        /* Input analysis */

        bool rotate_left=false, rotate_right=false;

        key_event_t e;
        bool stop_playing=false;

        while ((e = pollevent()).type != KEYEV_NONE) {
            if (e.type == KEYEV_DOWN && e.key == KEY_MENU)
                gint_osmenu();
            if (e.type == KEYEV_DOWN && e.key == KEY_EXIT)
                stop_playing = true;
        }
        if (keydown(KEY_7)) {
            rotate_left = true;
        }
        if (keydown(KEY_0)) {
            rotate_right = true;
        }
        if (stop_playing)
            break;

        /* Level transitions */

        // End of level
        if(state == State_Playing &&
                game.time >= lv->blocks[lv->block_count-1].time + 5) {
            bool changed_episode = false;

            current_level++;
            if(current_level >= episodes[current_episode].level_count) {
                current_episode++;
                current_level = 0;
                changed_episode = true;
            }
            if(current_episode >= episode_count)
                break;

            load_level(&game, episodes[current_episode].levels[current_level]);

            if(changed_episode) {
                state = State_EpisodeTransition;
                game.time_episode_transition = 0;
                game.forced_player_rota = -game.player_rota / 1.0;
                game.time = -13.0;
                continue;
            }
            else {
                state = State_Transition;
                game.time_transition = 0;
                game.forced_player_rota = -game.player_rota / 1.0;
                continue;
            }
        }

        // End of rewind
        if(state == State_Rewind && game.time <= -5.0) {
            game.time = -5.0;
            game.player_rota = 0.0;
            game.forced_player_rota = 0.0;
            state = State_Playing;
            continue;
        }

        /* Physics */

        if(state == State_Playing && player_collision(&game)) {
            state = State_Dead;
            game.time_dead = 0;
            continue;
        }

        for(int i = 0; i < game.rect_count; i++) {
            rect_physics(&game.rects[i], game.rects[i].meta, game.time);
        }

        if(game.forced_player_rota)
            game.player_rota += game.forced_player_rota * dt;
        else {
            if(state == State_Playing && rotate_left)
                game.player_rota += 1.57 * dt;
            if(state == State_Playing && rotate_right)
                game.player_rota -= 1.57 * dt;
        }
        if(game.player_rota > M_PI)
            game.player_rota -= 2 * M_PI;
        if(game.player_rota < -M_PI)
            game.player_rota += 2 * M_PI;

        /* Rendering */

        dclear(C_BLACK);

        // Cool animated background
        for(int i = 0; i < 10; i++) {
            int x=DWIDTH+2-40*i-20, y=20;

            rect_t rect = { .x=x, .y=y, .w=20, .h=20, .r=1.0*game.time };
            drectoid(&rect, 0, C_RGB(4, 4, 4));
        }
        for(int j = 1; j < 6; j++) {
            int lines = (j == 5) ? 24 : 40;
            dma_memcpy(gint_vram + 396*(40*j), gint_vram, 396*2*lines);
        }

        if(state == State_Playing && game.time < -4.0 && lv->message) {
            int x = DWIDTH/2 + 8 * (strcount(lv->message, '\n') + 1);

            /* Split at newlines */
            char const *str = lv->message;
            while(*str) {
                char const *end = strchrnul(str, '\n');
                duet_text_opt(x, DHEIGHT/2, C_WHITE, C_NONE, DTEXT_CENTER,
                    DTEXT_MIDDLE, str, end - str);
                x -= 16;
                str = end + (*end != 0);
            }
        }

        if(state == State_EpisodeTransition) {
            duet_text_opt(DWIDTH/2, DHEIGHT/2, C_WHITE, C_NONE, DTEXT_CENTER,
                DTEXT_MIDDLE, episodes[current_episode].name, -1);
        }

        render_player(PLAYER_X, DHEIGHT/2, game.player_rota);

        float extra_size = 1 + sinf(6 * game.time);
        for(int i = 0; i < game.rect_count; i++)
            drectoid(&game.rects[i], extra_size, C_WHITE);
        dupdate();
    }

    timer_stop(timer);
}

int main(void)
{
    __printf_enable_fp();

    /* Azur trickz for less tearing */
    r61524_set(0x010, 0x0010);

    int current_episode = 0;
    int current_level = 0;

    while(1) {
        main_menu(&current_episode, &current_level);
        game_loop(current_episode, current_level);
    }

    return 1;
}
